// ParticleSystem.cpp: implementation of the CParticleSystem class.
//
//////////////////////////////////////////////////////////////////////

#include "ParticleSystem.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CParticleSystem::CParticleSystem(Vector position)
{
	Position=position;
	Smoothstart=false;
	Smoothend=false;
	Smoothing=false;
	NumParticles=0;
	State=CREATED;
}

CParticleSystem::~CParticleSystem()
{

}

void CParticleSystem::SmoothStart(void) {
	Smoothstart=true;
}

void CParticleSystem::SmoothEnding(void) {
	Smoothend=true;
}

void CParticleSystem::DestroyParticle(std::list<CParticle>::iterator index) {
	(*index).life=0;
	(*index).alive=false;
}

// -------------------------------------------------------------------

CParticleRain::CParticleRain(Vector position):CParticleSystem(position) {
	Params.CheckCollision=false;
	Params.DropSpeed=0.05f;
	Params.NumOfDrops=100;
	Params.RainPlaneSize=2.0f;
	Params.Step=0.30f;
	Step=0;
	Position=position;
}

void CParticleRain::CreateParticle(void) {
	// New particle coordinates should be randomized in the XZ plane)

	CParticle Particle;
	RecycleParticle(&Particle);
	Particles.push_back(Particle);
	++NumParticles;
}

void CParticleRain::RecycleParticle(CParticle* position) {
	(*position).life=100;
	(*position).alive=true;
	float randomx = ((float) rand())/RAND_MAX;
	float randomy = ((float) rand())/RAND_MAX;
	float randomz = ((float) rand())/RAND_MAX;
	float halfsize=Params.RainPlaneSize/2;
	float positionx;
	float positiony;
	float positionz;

	positionx=-halfsize+(Params.RainPlaneSize*randomx);
	positiony=-1.0f+Position.y+randomy;
	positionz=-halfsize+(Params.RainPlaneSize*randomz);

	Vector pos(positionx, positiony, positionz);
	Vector direction(0.0f,-Params.DropSpeed,0.0f);
	(*position).position=pos;
	(*position).prev_position=pos;
	(*position).direction=direction*Params.DropSpeed;
	if (!(*position).alive) {
		(*position).alive=true;
	}
	(*position).life=100;
}

// Discover why the heck I can't delete a particle

bool CParticleRain::Update(float delta) {
	switch (State) {
	case CREATED:
	case STARTING:
		if (Smoothstart) {
			if (NumParticles < Params.NumOfDrops) {
				if (Step>=Params.Step) {
					CreateParticle();
					Step=0;
				} else {
					Step+=delta;
				}
			} else {
				State=STARTED;
			}
		} else {
			for (int loop=0; loop < Params.NumOfDrops; loop++) {
				CreateParticle();
			}
			State=STARTED;
		};
	case STARTED:
	case STOPPING: {
					std::list<CParticle>::iterator loop = Particles.begin();
					glDisable(GL_TEXTURE_2D);
					glBegin(GL_LINES);
					while (loop != Particles.end()) { // Update all particles
						if ((*loop).alive) {
							(*loop).prev_position=(*loop).position;
							(*loop).position=(*loop).position+((*loop).direction)*delta;
							(*loop).prev_position=(*loop).position;
							(*loop).prev_position.y-=0.1f;

							glColor3f(0.5f, 0.5f, 0.5f);
								glVertex3f((*loop).prev_position.x, (*loop).prev_position.y, (*loop).prev_position.z);
								glVertex3f((*loop).position.x, (*loop).position.y, (*loop).position.z);

							if ((*loop).position.y <= 0.0f) { 
								if (State!=STOPPING) {
									RecycleParticle(&(*loop));
								} else {
									DestroyParticle(loop);
								}
								if ((State==STOPPING) && (NumParticles==0)) {
									State=STOPPED;
									return false;
								}
							}
						}

						++loop;
					}
					glEnd();
					glEnable(GL_TEXTURE_2D);
					break;
				   }
	case STOPPED: return false; 
					break;
	}
	return true;
}

void CParticleRain::SetDefaults(RainParams params) {
	Params=params;
}

//-------------------------------------------------------------------------------------

CLightning::CLightning(Vector s, Vector e, float amplitude, int branches) {
	LightningCylinder=gluNewQuadric();
	gluQuadricNormals(LightningCylinder, GLU_SMOOTH);
	gluQuadricOrientation(LightningCylinder,GLU_OUTSIDE);
	gluQuadricTexture(LightningCylinder, GL_TRUE);
	Start=s;
	End=e;
	Amplitude=amplitude;
	NumBranches=branches;
	CallLightning();
}

CLightning::~CLightning() {
	gluDeleteQuadric(LightningCylinder);
	std::list<CLightning*>::iterator loop;
	for (loop = Trunks.begin(); loop!=Trunks.end(); loop++) 
		if (*loop) delete (*loop);
}

void CLightning::CallLightning() {
	CBranch ray;
	ray.origin=Start;
	ray.destination=End;
	CreateBranch(ray, NumBranches);
}

void CLightning::CreateBranch(CBranch branches, int num) {
	if (num==0) {
		Branches.push_back(branches);
		return;
	}
	float pertx, pertz;
	float half;
	half=Amplitude/2;

	pertx=-half+(half*rand()/RAND_MAX);
	pertz=-half+(half*rand()/RAND_MAX);

	Vector midpoint;
	midpoint=(branches.origin+branches.destination)*0.5f;
	midpoint.x+=pertx;
	midpoint.z+=pertz;
	CBranch branch1;
	branch1.origin=branches.origin;
	branch1.destination=midpoint;
	CBranch branch2;
	branch2.origin=midpoint;
	branch2.destination=branches.destination;
	float chewie=((float) rand()/(float) RAND_MAX);
	if ((chewie<0.1f)) {
		if (midpoint.y<=1.0f) {
		float px = -half+(half*rand()/RAND_MAX);
		float py = -half+(half*rand()/RAND_MAX);
		float pz = -half+(half*rand()/RAND_MAX);
		Vector pert2(px, py, pz);
		Vector newdest(midpoint+pert2);
		newdest.y=0.0f;
		CLightning* lightning = new CLightning(midpoint, newdest, Amplitude, 3);
		Trunks.push_back(lightning);
		}
	}

	CreateBranch(branch1, num-1);
	CreateBranch(branch2, num-1);
}

void CLightning::Update() {
	glColor3f(1.0f,1.0f,1.0f);
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_LIGHTING);
	glBegin(GL_LINES);
	std::list<CBranch>::iterator loop = Branches.begin();
	std::list<CBranch>::iterator loop2 = Branches.begin();
	std::list<CLightning*>::iterator loopl = Trunks.begin();

	while (loop!=Branches.end()) {
		++loop2;
		float length=((*loop).destination-(*loop).origin).Length();
		glColor3f(1.0f,1.0f,1.0f);
	
		glVertex3f((*loop).origin.x, (*loop).origin.y, (*loop).origin.z);
		glVertex3f((*loop).destination.x, (*loop).destination.y, (*loop).destination.z);
		++loop;
	}
	glEnd();
	for (loopl = Trunks.begin(); loopl != Trunks.end() ; loopl++)
		if (*loopl)(*loopl)->Update();

	glEnable(GL_TEXTURE_2D);
	glEnable(GL_LIGHTING);
}

//--------------------------------------------------------------------

CParticleLightning::CParticleLightning(Vector position):CParticleSystem(position) {
	Params.MinInterval=4000;
	Params.MaxInterval=10000;
	Params.Permanency=500;
	Params.MaxSubdivisions=6;
	Params.Amplitude=0.08f;
	Params.MaxBranches=0;
	Params.LightningPlaneSize=2.0f;
	LightningExists=false;
	ShouldStop=false;
	Acumulator=0.0f;
	Lightning=0;
}

CParticleLightning::~CParticleLightning() {
	if (Lightning) delete Lightning;
}

void CParticleLightning::CreateParticle(void) {
}

CLightning* CParticleLightning::CreateLightning(Vector startpoint, Vector endpoint) {
	LightningExists=true;
	return (new CLightning(startpoint, endpoint, Params.Amplitude, Params.MaxSubdivisions));
}

void CParticleLightning::SetDefaults(LightningParams params) {
	Params=params;
}

void CParticleLightning::Stop(void) {
	ShouldStop=true;
}

bool CParticleLightning::Update(float delta) {
	if (ShouldStop) return false;
	switch(State) {
	case CREATED: State=STARTED; break;
	case STARTED: {
		if (!LightningExists) {
			if (Trigger==0.0f) {
				Trigger=Params.MinInterval+(Params.MaxInterval-Params.MinInterval)*rand()/RAND_MAX;
			} else {
				if (Trigger<=Acumulator) {
					float randomx = ((float) rand())/RAND_MAX;
					float randomz = ((float) rand())/RAND_MAX;
					float halfsize=Params.LightningPlaneSize/2;
					float positionx;
					float positiony;
					float positionz;

					positionx=-halfsize+(Params.LightningPlaneSize*randomx);
					positiony=Position.y;
					positionz=-halfsize+(Params.LightningPlaneSize*randomz);

					Vector start(positionx, positiony, positionz);
					randomx= ((float) rand())/RAND_MAX;
					randomz= ((float) rand())/RAND_MAX;
					positionx=-halfsize+(Params.LightningPlaneSize*randomx);
					positionz=-halfsize+(Params.LightningPlaneSize*randomz);
					Vector end(positionx, 0.0f, positionz);

					Lightning=CreateLightning(start, end);
					LightningExists=true;
					LightDiffuse[0]=1.0f;
					LightDiffuse[1]=1.0f;
					LightDiffuse[2]=1.0f;
					LightDiffuse[3]=1.0f;
					glLightfv(GL_LIGHT1,GL_DIFFUSE,LightDiffuse);
					PlaySound("Thundr3.wav", NULL, SND_ASYNC);
					//PlaySound("rain.wav", NULL, SND_ASYNC | SND_LOOP);
	
					Acumulator=0.0f;
				} else {
					Acumulator+=delta;
				}
			}
		} else {
			if (Acumulator>=Params.Permanency) {
				delete Lightning;
				LightDiffuse[0]=0.2f;
				LightDiffuse[1]=0.2f;
				LightDiffuse[2]=0.2f;
				LightDiffuse[3]=0.2f;
				glLightfv(GL_LIGHT1,GL_DIFFUSE,LightDiffuse);
				
				LightningExists=false;
				Acumulator=0.0f;
				Trigger=0.0f;
			} else {
				Lightning->Update();
				Acumulator+=delta;
			}
		}
				  } // case
	}
	return true;
}

//-----------------------------------------------------------------------

CParticleSmoke::CParticleSmoke(Vector position, int index):CParticleSystem(position) {
	Params.Speed=0.1f;
	Params.Spread=0.00008f;
	Params.Speed=0.0001f;
	Params.MaxParticles=400;
	Params.Step=0.10f;
	Params.Gravity=0.0f;
	Params.AtTime=5;
	Params.MaxHeight=1.0f;
	Params.size=0.07f;
	Step=0;
	Index=index;
	running=false;
	//LoadTGA(&Smoke[0],"blackoak2.tga");
}

void CParticleSmoke::RecycleParticle(CParticle* position) {
	(*position).alive=true;
	(*position).life=1.0f;
	(*position).position=Position;
	(*position).prev_position=Position;
	(*position).direction.x=-(Params.Spread/2)+(Params.Spread*rand()/RAND_MAX);
	(*position).direction.z=-(Params.Spread/2)+(Params.Spread*rand()/RAND_MAX);
	(*position).direction.y=Params.Speed;
	//float randomz = ((float) rand())/RAND_MAX;
	//positionx=-halfsize+(Params.RainPlaneSize*randomx);

//
}

void CParticleSmoke::CreateParticle() {
	CParticle Particle;
	RecycleParticle(&Particle);

	Particles.push_back(Particle);

	++NumParticles;
}

CParticleSmoke::~CParticleSmoke() {
}

void CParticleSmoke::TransposeMatrix(GLfloat* m) {
   GLfloat temp[16];
   temp[0] = m[0]; temp[1] = m[4]; temp[2] = m[8]; temp[3] = m[12];
   temp[4] = m[1]; temp[5] = m[5]; temp[6] = m[9]; temp[7] = m[13];
   temp[8] = m[2]; temp[9] = m[6]; temp[10] = m[10]; temp[11] = m[14];
   temp[12] = m[3]; temp[13] = m[7]; temp[14] = m[11]; temp[15] = m[15];
   for (int loop=0; loop<16; loop++) 
      m[loop]=temp[loop];
};

bool CParticleSmoke::Update(float delta) {
	if (!running) return true;
	GLfloat mat[16];

	glGetFloatv(GL_MODELVIEW_MATRIX, mat);//pode estar cp
	TransposeMatrix(mat);//pode estar cp
	
	glDisable(GL_LIGHTING);
	
	//glBindTexture(GL_TEXTURE_2D, Smoke[0].texID);
	glBindTexture(GL_TEXTURE_2D, terreno->Smoke[Index].texID);
	
	switch (State) {
	case CREATED:
	case STARTING:
			if (NumParticles < Params.MaxParticles) {
				if (Step>=Params.Step) {
					for (int loop=0; loop < Params.AtTime; loop++)
					CreateParticle();
					Step=0;
				} else {
					Step+=delta;
				}
			} else {
				State=STARTED;
			}
	case STARTED:
	case STOPPING: {
					glEnable(GL_BLEND);
					glBlendFunc(GL_SRC_ALPHA, GL_SRC_COLOR);

					glEnable(GL_ALPHA_TEST);
					glAlphaFunc(GL_GREATER, 0.1f);
					std::list<CParticle>::iterator loop = Particles.begin();
					//glDisable(GL_LIGHTING);
					float size=Params.size;
					while (loop != Particles.end()) { // Update all particles
						if ((*loop).alive) {
							(*loop).prev_position=(*loop).position*delta*000.1f;
							(*loop).position=(*loop).position+((*loop).direction)*delta;
							(*loop).direction.y-=Params.Gravity;
							//(*loop).life=-0.0000001f*delta;

							//glColor4f(1.0f,1.0f,1.0f,(*loop).life);
							glColor4f(1.0f,1.0f,1.0f,0.2f);
							glBegin(GL_QUADS);
							glTexCoord2f(0.0f,0.0f);
							glVertex3f((*loop).position.x+(((-mat[4])-mat[0])*size), (*loop).position.y+(((-mat[5])-mat[1])*size)
					                    ,(*loop).position.z+(((-mat[6])-mat[2])*size));
							glTexCoord2f(1.0f,0.0f);              
							glVertex3f((*loop).position.x+(((mat[4])-mat[0])*size), (*loop).position.y+(((mat[5])-mat[1])*size)
									,(*loop).position.z+(((mat[6])-mat[2])*size));
							glTexCoord2f(1.0f,1.0f);              
							glVertex3f((*loop).position.x+(((mat[4])+mat[0])*size), (*loop).position.y+(((mat[5])+mat[1])*size)
									,(*loop).position.z+(((mat[6])+mat[2])*size));
							glTexCoord2f(0.0f,1.0f);              
							glVertex3f((*loop).position.x+(((-mat[4])+mat[0])*size), (*loop).position.y+(((-mat[5])+mat[1])*size)
									,(*loop).position.z+(((-mat[6])+mat[2])*size));   
							glEnd();


							if (((*loop).position.y >= Params.MaxHeight) || ((*loop).life<=0.0f) || ((*loop).position.y<=-Params.MaxHeight)) { 
								if (State!=STOPPING) {
									RecycleParticle(&(*loop));
								} else {
									DestroyParticle(loop);
								}
								if ((State==STOPPING) && (NumParticles==0)) {
									State=STOPPED;
									return false;
								}
							}
						}

						++loop;
					}
					glDisable(GL_BLEND);
					glDisable(GL_ALPHA_TEST);
					//glEnable(GL_LIGHTING);
					//glDisable(GL_BLEND);
					//std::cout << "Exited form the UPDATE loop." << std::endl;
					break;
				   }
	case STOPPED: return false; 
				  break;
	}
	glEnable(GL_LIGHTING);
	//std::cout << "Returning TRUE. " << std::endl;
	return true;
}

bool CParticleSmoke::LoadTGA(Texture *texture, char *filename)
{    
   GLubyte		TGAheader[12]={0,0,2,0,0,0,0,0,0,0,0,0};
   GLubyte		TGAcompare[12];
   GLubyte		header[6];
   GLuint		bytesPerPixel;
   GLuint		imageSize;
   GLuint		temp;
   GLuint		type=GL_RGBA;		
   FILE *file = fopen(filename, "rb");
   if(	file==NULL ||
      fread(TGAcompare,1,sizeof(TGAcompare),file)!=sizeof(TGAcompare) ||
      memcmp(TGAheader,TGAcompare,sizeof(TGAheader))!=0		||
      fread(header,1,sizeof(header),file)!=sizeof(header))
   {
      if (file == NULL)
         return false;
      else {
         fclose(file);
         return false;
      }
   }
   texture->width  = header[1] * 256 + header[0];
   texture->height = header[3] * 256 + header[2];
   if(	texture->width	<=0	||
   texture->height	<=0	||
   (header[4]!=24 && header[4]!=32))
   {
      fclose(file);
      return false;
   }
   texture->bpp	= header[4];
   bytesPerPixel	= texture->bpp/8;
   imageSize	= texture->width*texture->height*bytesPerPixel;
   texture->imageData=(GLubyte *)malloc(imageSize);
   if(	texture->imageData==NULL ||
   fread(texture->imageData, 1, imageSize, file)!=imageSize)
   {
      if(texture->imageData!=NULL)
         free(texture->imageData);
      fclose(file);
      return false;
   }
   for(GLuint i=0; i<int(imageSize); i+=bytesPerPixel) {
      temp=texture->imageData[i];
      texture->imageData[i] = texture->imageData[i + 2];
      texture->imageData[i + 2] = temp;
   }

   fclose (file);		

   glGenTextures(1, &texture[0].texID);
   glBindTexture(GL_TEXTURE_2D, texture[0].texID);
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
   glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
   if (texture[0].bpp==24)
   {
      type=GL_RGB;
   } 
   glTexImage2D(GL_TEXTURE_2D, 0, type, texture[0].width, texture[0].height, 0, type, GL_UNSIGNED_BYTE, texture[0].imageData);
   return true;
}

void CParticleSmoke::SetDefaults(SmokeParams params) {
	Params=params;
}