// ParticleSystem.h: interface for the CParticleSystem class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_PARTICLESYSTEM_H__38ADDD66_AD3E_48F8_B280_C5C49DEB7089__INCLUDED_)
#define AFX_PARTICLESYSTEM_H__38ADDD66_AD3E_48F8_B280_C5C49DEB7089__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include <cstdlib>
#include <list>
#include <windows.h>
#include <iostream>
#include <GL/gl.h>
#include <GL/glu.h>

#include "Particle.h"
#include "mathclass.h"
#include "Ray.h"
#include "Texture.h"
#include "terrain.h"

class Terrain;

enum PARTICLESTATE { CREATED, STARTING, STARTED, STOPPING, STOPPED };

class CParticleSystem  
{
protected:
	std::list<CParticle> Particles;
	Vector Position;
	bool Smoothstart;
	bool Smoothend;
	bool Smoothing;
	PARTICLESTATE State;
	int NumParticles;
public:
	Terrain* terreno;
	virtual void SmoothStart(void);
	virtual void SmoothEnding(void); // Only for "permanent" systems

	virtual bool Update(float delta)=0;
	virtual void CreateParticle(void)=0;
	virtual void DestroyParticle(std::list<CParticle>::iterator index);

	CParticleSystem(Vector position);
	virtual ~CParticleSystem();
};

// -------------------------------------------------------------------

struct RainParams {
	int NumOfDrops;
	float DropSpeed;
	bool CheckCollision;
	float RainPlaneSize;
	float Step;
};

class CParticleRain : public CParticleSystem {
protected:
	RainParams Params;
	float Step;
public:
	virtual void CreateParticle(void);
	void CParticleRain::RecycleParticle(CParticle* position); 
	virtual bool Update(float delta);
	void SetDefaults(RainParams params);

	CParticleRain(Vector position);
};

class CBranch {
public:
	Vector origin;
	Vector destination;
};

class CLightning {
private:
	Vector Start;
	Vector End;
	std::list<CBranch> Branches;
	std::list<CLightning*> Trunks;
	float Amplitude;
	int NumBranches;
	GLUquadricObj* LightningCylinder;
public:
	void CreateBranch(CBranch branches, int num);
	void CallLightning();
	void Update(void);

	CLightning(Vector s, Vector e, float amplitude, int branches);
	virtual ~CLightning();
};

struct LightningParams {
	float MinInterval;
	float MaxInterval;
	float Permanency;
	int MaxSubdivisions;
	int MaxBranches;
	float Amplitude;
	float LightningPlaneSize;
};

class CParticleLightning : public CParticleSystem {
protected:
	LightningParams Params;
	bool ShouldStop;
	bool LightningExists;
	GLfloat LightDiffuse[4];
	float Acumulator;
	float Trigger;
	CLightning* Lightning;
public:
	void SetDefaults(LightningParams params);
	CLightning* CreateLightning(Vector startpoint, Vector endpoint);
	virtual void CreateParticle(void);
	virtual bool Update(float delta);
	void Stop(void);

	CParticleLightning(Vector position);
	virtual ~CParticleLightning();
};

struct SmokeParams {
	double MaxLife;
	double Spread;
	double Speed;
	double MaxParticles;
	double Step;
	double Gravity;
	float MaxHeight;
	float size;
	int AtTime;
};

typedef	struct	{
	GLubyte	*imageData;
	GLuint	bpp;
	GLuint	width;
	GLuint	height;
	GLuint	texID;
} Texture;


class CParticleSmoke : public CParticleSystem {
protected:
	SmokeParams Params;
	float Step;
	Texture Smoke[1];
	CTexture textura;
	int Index;
	bool running;
public:
	void Start() {
		running=true;
	};
	void Stop() {
		State=STOPPING;
	};
	void SetDefaults(SmokeParams params);
	virtual void CreateParticle(void);
	virtual bool Update(float delta);
	void RecycleParticle(CParticle* position);
	void TransposeMatrix(GLfloat* m);
	bool LoadTGA(Texture *treetex, char *filename);
	CParticleSmoke(Vector position, int index);
	virtual ~CParticleSmoke();
};

#endif // !defined(AFX_PARTICLESYSTEM_H__38ADDD66_AD3E_48F8_B280_C5C49DEB7089__INCLUDED_)
